;;; mailcap-mode.el --- Major mode for editing mailcap files -*- lexical-binding: t -*-

;; Copyright (C) 2017  Mark Oteiza <mvoteiza@udel.edu>

;; Author: Mark Oteiza <mvoteiza@udel.edu>
;; Created: 02 September 2017
;; Keywords: wp, mail, multimedia,

;; This program is free software; you can redistribute it and/or
;; modify it under the terms of the GNU General Public License
;; as published by the Free Software Foundation; either version 3
;; of the License, or (at your option) any later version.

;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program. If not, see <http://www.gnu.org/licenses/>.

;;; Commentary:

;; Major mode for editing mailcap files.

;; References:

;; https://tools.ietf.org/html/rfc1524
;; https://tools.ietf.org/html/rfc1521
;; https://www.iana.org/assignments/media-types/media-types.xhtml
;; https://www.ch.ic.ac.uk/chemime/

;;; Code:

(require 'conf-mode)

(defun mailcap-entry-start-p ()
  "Return non-nil if the current line is the first in a mailcap entry."
  (let ((flag t))
    (save-excursion
      (while (and (zerop (forward-line -1))
                  (save-excursion
                    (goto-char (line-end-position))
                    (eq (preceding-char) ?\\))
                  (skip-chars-forward " \t")
                  (setq flag (eq (following-char) ?#)))))
    flag))

(defun mailcap-comment-anchored-matcher (limit)
  "Matcher for comments, which must be of the form ^#[...].
See https://tools.ietf.org/html/rfc1524 for details."
  (let ((pos (car (match-data))) res)
    (when (mailcap-entry-start-p)
      (goto-char limit)
      (prog1 (setq res (list pos (point-marker)))
        (message "%S" res)
        (set-match-data res)))))

(defun mailcap-typefield-anchored-matcher (limit)
  "Matcher for the MIME type form TYPE/SUBTYPE.
See https://tools.ietf.org/html/rfc1521 for details."
  (let ((pos (car (match-data)))
        (token "[:alnum:]!#$%&'*+-.^_`{|}~")
        res)
    (when (and (mailcap-entry-start-p)
               (< 0 (skip-chars-forward token limit))
               (= (following-char) ?\/))
      (forward-char 1)
      (when (and (< 0 (skip-chars-forward token limit))
                 (skip-chars-forward " \t" limit)
                 (= (following-char) ?\;))
        (prog1 (setq res (list pos (point-marker)))
          (set-match-data res))))))

(defun mailcap-limit-field ()
  (when (not (eolp))
    (let ((limit (line-end-position)))
      (save-excursion
        (save-match-data
          (when (search-forward ";" limit t 1)
            (point)))))))

(defun mailcap-field-anchored-matcher (limit)
  "Matcher for the optional field names in the form FIELD[=ARG].
See https://tools.ietf.org/html/rfc1524 for details."
  (let ((fields
         (eval-when-compile
           (regexp-opt
            '("compose" "composetyped" "edit" "print" "test" "needsterminal"
              "copiousoutput" "description" "textualnewlines" "x11-bitmap"
              "nametemplate")
            'symbols)))
        res)
    (let ((case-fold-search t))
      (setq res (re-search-forward fields limit t 1)))
    (when (null res)
      (re-search-forward "x-[[:alnum:]!#$%&'*+-.^_`{|}~]+" limit t 1))
    (when res
      (prog1 (setq res (list (car (match-data)) (point-marker)))
        (set-match-data res)))))


;; Font lock multiline

(defvar font-lock-beg)
(defvar font-lock-end)

(defun mailcap-entry-start-index ()
  "Return a number for `forward-line' to the beginning of a construct."
  (let ((count 0))
    (save-excursion
      (while (and (zerop (forward-line -1))
                  (save-excursion
                    (goto-char (line-end-position))
                    (eq (preceding-char) ?\\))
                  (skip-chars-forward " \t")
                  (/= (following-char) ?#))
        (cl-decf count)))
    count))

(defun mailcap-entry-end-index ()
  "Return the number for `forward-line' to the end of construct."
  (let ((count 0))
    (save-excursion
      (while (and (save-excursion
                    (goto-char (line-end-position))
                    (eq (preceding-char) ?\\))
                  (skip-chars-forward " \t")
                  (/= (following-char) ?#)
                  (zerop (forward-line)))
        (cl-incf count)))
    count))

(defun mailcap-font-lock-extend-region ()
  "Function for `font-lock-extend-region-functions'."
  ;; This always returns non-nil--why does this work?
  (let (idx1 idx2 changed)
    (goto-char font-lock-beg)
    (when (setq idx1 (mailcap-entry-start-index))
      (forward-line idx1)
      (setq font-lock-beg (point) changed t))
    (goto-char font-lock-end)
    (when (setq idx2 (mailcap-entry-end-index))
      (forward-line idx2)
      (setq font-lock-end (line-end-position) changed t))
    changed))

(defvar mailcap-mode-syntax-table
  (let ((table (make-syntax-table)))
    (modify-syntax-entry ?\\ "/" table)
    (modify-syntax-entry ?% "'" table)
    (modify-syntax-entry ?\; "." table)
    (modify-syntax-entry ?# "<" table)
    table)
  "Syntax table used in `mailcap-mode' buffers.")

(defconst mailcap-font-lock-keywords-1
  '(("^#"
     (mailcap-comment-anchored-matcher nil nil (0 'font-lock-comment-face)))
    ("^[aAcCfFeEiImMtTvV]"              ; type/subtype;
     (mailcap-typefield-anchored-matcher nil nil (0 'font-lock-keyword-face))))
  "Minimal keywords to highlight in `mailcap-mode'.")

(defconst mailcap-font-lock-keywords-2
  `(,@mailcap-font-lock-keywords-1
    ("\\\\$" 0 'font-lock-warning-face) ; line break
    ("%{\\(?:boundary\\|charset\\|media-type\\)}"
     0 'font-lock-variable-name-face)
    ("%[stFn]" 0 'font-lock-constant-face)
    (";[ \t]*?"
     (mailcap-field-anchored-matcher
      (mailcap-limit-field) nil (0 'font-lock-builtin-face))))
  "Accurate normal keywords to highlight in `mailcap-mode'.")

(defvar mailcap-font-lock-keywords 'mailcap-font-lock-keywords-1
  "Default expressions to highlight in `mailcap-mode'.")

;;;###autoload (add-to-list 'auto-mode-alist '("\\.mailcap\\'" . mailcap-mode))

(define-derived-mode mailcap-mode conf-mode "Mailcap"
  "Major mode for editing mailcap files."
  (set-keymap-parent mailcap-mode-map nil)
  (conf-mode-initialize "#")
  (setq-local comment-style 'plain)
  (add-hook 'font-lock-extend-region-functions
            'mailcap-font-lock-extend-region nil t)
  (setq font-lock-defaults
        '((mailcap-font-lock-keywords
           mailcap-font-lock-keywords-1 mailcap-font-lock-keywords-2))))

(provide 'mailcap-mode)

;;; mailcap-mode.el ends here
